local libName, lib = ...
local s = SpellFlashAddon

BittensSpellFlashLibrary = {}
local c = BittensSpellFlashLibrary

BINDING_HEADER_BITTENS_SPELLFLASH = select(2, GetAddOnInfo(libName))
BINDING_NAME_BITTENS_SPELLFLASH_DEBUGGING
    = lib.Localize["Print Debugging Info"]

local GetTime = GetTime
local UnitGroupRolesAssigned = UnitGroupRolesAssigned
local UnitIsUnit = UnitIsUnit
local UnitThreatSituation = UnitThreatSituation
local math = math
local pairs = pairs
local print = print
local select = select
local string = string
local table = table
local type = type
local unpack = unpack

c.Debugging = false

function c.Init(a)
    lib.A = a
end

function c.GetCurrentRotation(a)
    for _, rotation in pairs(a.Rotations) do
        if (rotation.CheckFirst == nil or rotation.CheckFirst())
            and rotation.Spec == s.TalentMastery()
            and (rotation.OffSwitch == nil
                or not a.GetConfig(rotation.OffSwitch))
            and (rotation.Form == nil or s.Form(c.GetID(rotation.Form))) then
            
            return rotation
        end
    end
    return nil
end

function c.Flash(a)
    local rotation = c.GetCurrentRotation(a)
    if rotation ~= nil then
        c.HideHilighting = true
        c.Init(a)
        if rotation.Function then -- depricated, do not use, will be phased out.
            rotation.Function()
        else
            if rotation.FlashAlways then
                rotation.FlashAlways()
            end
            if s.InCombat() then
                if rotation.FlashInCombat then
                    rotation.FlashInCombat()
                end
            else
                if rotation.FlashOutOfCombat then
                    rotation.FlashOutOfCombat()
                end
            end
        end
    end
end

-- This function is depricated.  It will be phased out.  Use c.Flash()
function c.Run(...)
    for i=1,select("#", ...) do
        local runInfo = select(i, ...)
        
        local func = runInfo[1]
        local spec = runInfo[2]
        local offSwitch = runInfo[3]
        local form = runInfo[4]
        
        if s.TalentMastery(spec) 
            and not lib.A.GetConfig(offSwitch) 
            and (form == nil or s.Form(c.GetID(form))) then
            
            c.HideHilighting = true
            func()
            return
        end
    end
    c.HideHilighting = false
end

function c.GetSpell(name)
	local spell = lib.A.Spells[name]
	if spell == nil then
		spell = c.AddSpell(name)
	end
	return spell
end

function c.RotateCooldowns(uncontrolledBuffs, ...)
    for _, buff in pairs(uncontrolledBuffs) do
        if c.HasBuff(buff) then
            return
        end
    end
    for i = 1, select("#", ...) do
        buff = select(i, ...)
        local spell
        if lib.A.SpellIDs then
        	spell = c.GetSpell(buff)
        else
        	spell = lib.A.spells[buff]
        end
        if spell and spell.Buff then
            buff = spell.Buff
        end
        if c.GetBuffDuration(buff, spell.NoGCD) > (spell.EarlyRefresh or 0) then
            return
        end
    end
    if lib.A.SpellIDs then
	    for i = 1, select("#", ...) do
	        buff = select(i, ...)
	        if s.CheckThenFlash(c.GetSpell(buff)) then
	        	return true
	        end
	    end
    else
	    return lib.A.Flash(...)
	end
end

function c.HasBuff(name, noGCD)
    local duration = s.BuffDuration(c.GetID(name), "player")
    if duration == 0 then
        -- duration == 0 for permanent buffs
        return s.Buff(c.GetID(name), "player")
    else
        return duration > c.GetBusyTime(noGCD)
    end
end

function c.GetBuffDuration(name, noGCD)
    return math.max(
        s.BuffDuration(c.GetID(name), "player") - c.GetBusyTime(noGCD), 0)
end

function c.GetBuffStack(name)
    if c.HasBuff(name) then
        return s.BuffStack(c.GetID(name), "player")
    else
        return 0
    end
end

function c.HasMyDebuff(name, noGCD, durationRequired)
    if not durationRequired then
        durationRequired = 0
    end
    return c.GetMyDebuffDuration(name, noGCD) > durationRequired
end

function c.GetMyDebuffDuration(name, noGCD)
    return math.max(s.MyDebuffDuration(c.GetID(name)) - c.GetBusyTime(noGCD), 0)
end

function c.GetDebuffDuration(name, noGCD)
    return math.max(0, s.DebuffDuration(c.GetID(name)) - c.GetBusyTime(noGCD))
end

function c.GetCooldown(name)
    return math.max(0, s.SpellCooldown(c.GetID(name)) - c.GetBusyTime())
end

function c.GetIDs(...)
    local ids = {}
    for i = 1, select("#", ...) do
        local id = c.GetID(select(i, ...))
        if id then
            table.insert(ids, id)
        end
    end
    return ids
end

function c.GetID(name)
	if type(name) ~= "string" then
		return name
	end
	
	local id
	if lib.A.SpellIDs == nil then
	    local spell = lib.A.spells[name]
	    if spell ~= nil then
	        id = spell.ID
	    end
    else
    	id = lib.A.SpellIDs[name]
    end
    if id then
    	return id
   	else
  	  	print("No spell defined (or no ID attribute):", name)
  	end
end

function c.SelfBuffNeeded(name)
    if s.InCombat() then
        return not c.HasBuff(name)
    else
        return c.GetBuffDuration(name) < 5 * 60
    end
end

function c.RaidBuffNeeded(idTable)
    local duration = 0
    if not s.InCombat() then
        duration = 5 * 60
    end
    local flags = "raid|all|range"
    if idTable == c.INTELLECT_BUFFS then
        flags = flags .. "|mana"
    end
    return not s.Buff(idTable, flags, duration)
end

function c.Debug(tag, ...)
    if c.Debugging then
        return print("|cFF00FFFF["..tag.."]|r", ...)
    end
end

function c.IsCasting(name)
    return s.CurrentSpell(c.GetID(name))
end

function c.IsCastingOrInAir(name)
    return c.IsCasting(name) or s.SpellDelay(c.GetID(name))
end

function c.IsAuraPendingFor(name)
    return c.IsCastingOrInAir(name) or s.AuraDelay(c.GetID(name))
end

function c.IsTanking()
    local status = UnitThreatSituation("player", "target")
    return status == 2 or status == 3
end

function c.CheckFirstForTaunts(z)
    local primaryTarget = s.GetPrimaryThreatTarget()
    if not primaryTarget or UnitIsUnit(primaryTarget, "player") then
        return false
    end
    
    if UnitGroupRolesAssigned(primaryTarget) == "TANK" then
        z.FlashSize = s.FlashSizePercent() / 2
        z.FlashColor = "yellow"
    else
        z.FlashSize = nil
        z.FlashColor = "red"
    end
    return true
end

function c.WearingSet(number, name)
    local count = 0
    for slot, piece in pairs(lib.A.EquipmentSets[name]) do
        if s.Equipped(piece, slot) then
            count = count + 1
        end
    end
    return count >= number
end

function c.EstimateTravelTime(name)
    local spell = lib.A.spells[name]
    local id = spell.ID
    local travel, lastSeen = s.LastSpellTravelTime(id)
    if travel == nil then
        travel = spell.TravelTimeEstimate
        if travel == nil then
            print(name, "does not define TravelTimeEstimate")
            travel = 1
        end
        lastSeen = 0
    end
    
    if spell.SameTravelTimeAs ~= nil then
        for _, name in pairs(spell.SameTravelTimeAs) do
            local t, l = s.LastSpellTravelTime(c.GetID(name))
            if t ~= nil and l < lastSeen then
                travel = t
                lastSeen = l
            end
        end
    end
    return travel
end

local function incrementIfLanding(
    count, startDelay, endDelay, castDelay, travelTime)
    
    if castDelay == nil then
        return count
    end
    
    local landDelay = castDelay + travelTime
    if startDelay <= landDelay and landDelay < endDelay then
        return count + 1
    else
        return count
    end
end

function c.CountLandings(name, startDelay, endDelay, countNextCast)
    local count = 0
    local travel = c.EstimateTravelTime(name)
    
    -- in the air
    local foundCurrent = false
    local timesForAllTargets = s.SpellTravelStartTime(c.GetID(name), "all")
    if timesForAllTargets then
        for _, timesForOneTarget in pairs(timesForAllTargets) do
            for _, timesForOneLaunch in pairs(timesForOneTarget) do
                local estimated, actual = unpack(timesForOneLaunch)
                if estimated ~= nil then
	                if actual == nil then
	                    foundCurrent = true
	                end
	                count = incrementIfLanding(
	                    count,
	                    startDelay,
	                    endDelay,
	                    estimated - GetTime(),
	                    travel)
	            end
            end
        end
    end
    
    -- currently casting
    if c.IsCasting(name) and not foundCurrent then
        count = incrementIfLanding(
            count, startDelay, endDelay, c.GetCastTime(name), travel)
    end
    
    -- will have time to cast
    if countNextCast then
        count = incrementIfLanding(
            count,
            startDelay,
            endDelay,
            c.GetBusyTime() + c.GetCastTime(name),
            travel)
    end
    
    return count
end

function c.ShouldCastToRefresh(
    spellName, debuffName, earlyRefresh, willApplyDebuff, ...)
    
    local duration = s.MyDebuffDuration(c.GetID(debuffName))
    if duration == 0 then
        return willApplyDebuff and not c.IsAuraPendingFor(spellName)
    end
    
    if c.CountLandings(spellName, -3, duration) > 0 then
        return false
    end
    
    for i = 1, select("#", ...) do
        if c.CountLandings(select(i, ...), -3, duration) > 0 then
            return false
        end
    end
    
    local landing =
        c.GetBusyTime()
            + c.GetCastTime(spellName)
            + c.EstimateTravelTime(spellName)
    return duration - earlyRefresh + .1 < landing and landing < duration - .2
end

function c.HasTalent(name)
    return s.HasTalent(lib.A.TalentIDs[name])
end

function c.GetTalentRank(name)
    return s.TalentRank(lib.A.TalentIDs[name])
end

function c.HasGlyph(name)
	return s.HasGlyph(lib.A.GlyphIDs[name])
end

function c.RegisterAura(spellName, auraName)
    s.SetOtherAuras(c.GetID(spellName), c.GetID(auraName))
end

function c.GetCastTime(spellName)
    return s.CastTime(s.SpellName(c.GetID(spellName)))
end

function c.GetBusyTime(noGCD)
    local busy = s.GetCasting(nil, "player")
    
    local queued = c.GetQueuedInfo()
    if queued then
        busy = busy + s.CastTime(queued.Name)
    end
    
    if not noGCD then
        local gcd, _ = s.GlobalCooldown()
        busy = math.max(gcd, busy)
    end
    
    return busy
end

function c.GetPower(regen)
    local power = s.Power("player")
    local max = s.MaxPower("player")
    
--c.Debug("Lib", power)
    local info = c.GetCastingInfo()
    local gcd = s.GlobalCooldown()
    if info then
        local busy = math.max(s.GetCasting(nil, "player"), gcd)
        power = math.min(max, power + busy * regen)
        power = math.min(max, power - info.Cost)
--c.Debug("Lib", power, info.Name, "cast")
    else
        power = math.min(max, power + gcd * regen)
    end
    
    info = c.GetQueuedInfo()
    if info then
        power = math.min(max, power + s.CastTime(info.Name) * regen)
--c.Debug("Lib", power, info.Name, "regen", s.CastTime(info.Name))
        power = math.min(max, power - (info.Cost or s.SpellCost(info.Name)))
--c.Debug("Lib", power, info.Name, "queue")
    end
    
    return power
end

local function convertToIDs(attributes, key, ...)
	for i = 1, select("#", ...) do
		local pattern = select(i, ...)
		local _, last = string.find(key, pattern)
		if last == #key then
			local value = attributes[key]
			if type(value) == "table" then
				attributes[key] = c.GetIDs(value)
			else
				attributes[key] = c.GetID(value)
			end
		end
	end
end

function c.AddSpell(spellName, tag, attributes)
	local name = spellName
	if tag then
		name = name .. " " .. tag
	end
	if lib.A.Spells == nil then
		lib.A.Spells = {}
	elseif lib.A.Spells[name] then
		print("Warning:", name, "is already defined")
	end
	if attributes == nil then
		attributes = {}
	end
	for k, v in pairs(attributes) do
		convertToIDs(attributes, k, "ID", "Debuff%d*", "Buff%d*")
	end
	if attributes.ID == nil then
		attributes.ID = c.GetID(spellName)
	end
	lib.A.Spells[name] = attributes
	return attributes
end

function c.AddOptionalSpell(name, tag, attributes, color)
	local spell = c.AddSpell(name, tag, attributes)
	spell.Continue = true
	if color ~= nil then
		spell.FlashColor = color
	elseif spell.FlashColor == nil then
		spell.FlashColor = "yellow"
	end
	return spell
end

function c.AddInterrupt(name, tag, attributes)
	return c.AddOptionalSpell(name, tag, { Interrupt = true }, "aqua")
end

function c.AddTaunt(name, tag, attributes)
	return c.AddOptionalSpell(name, tag, { CheckFirst = c.CheckFirstForTaunts })
end

function c.CloneSpell(sourceName, tag, attributes)
	local sourceAttributes = lib.A.Spells[sourceName]
	if sourceAttributes == nil then
		print("Cannot clone", sourceName, "if it does not exist!")
		return
	end
	
    local spell = c.AddSpell(sourceName, tag, attributes)
    for k, v in pairs(sourceAttributes) do
    	if spell[k] == nil then
    		spell[k] = v
    	end
    end
    return spell
end

function c.PriorityFlash(...)
	local flashed = nil
	for i = 1, select("#", ...) do
		local name = select(i, ...)
		local spell = c.GetSpell(name)
		if s.CheckThenFlash(spell) then
			flashed = name
			if not spell.Continue then
				break
			end
		end
	end
	return flashed
end

function c.FlashAll(...)
	local flashed = false
	for i = 1, select("#", ...) do
		if s.CheckThenFlash(c.GetSpell(select(i, ...))) then
			flashed = true
		end
	end
	return flashed
end
